from collections.abc import MutableMapping
from collections import defaultdict
from typing import Any, DefaultDict, Iterator

from .queuedist import QueueDist


class SubscribableDict(MutableMapping):
    """Dict with async generator that yields changes
    """

    def __init__(self):
        self._map: DefaultDict[str, QueueDist] = defaultdict(QueueDist)

    def __getitem__(self, key) -> Any:
        if dp := self._map.get(key):
            return dp.get()
        else:
            raise KeyError

    def __setitem__(self, key, value) -> None:
        self._map[key].put(value)

    def __delitem__(self, key) -> None:
        dq = self._map.pop(key, None)
        if dq is None:
            return
        dq.close()

    def __iter__(self) -> Iterator[Any]:
        return iter(self._map)

    def __len__(self) -> int:
        return len(self._map)

    def close(self):
        """End gracefully"""
        for key in list(self):
            del self[key]

    def get(self, key, default=None):
        """The item for the key. The default if the key doesn't exist"""
        if dp := self._map.get(key):
            return dp.get()
        else:
            return default

    async def subscribe(self, key):
        dq = self._map[key]
        async for y in dq.subscribe():
            yield y



import asyncio
import itertools

import pytest

from nextline.state import Initialized
from nextline.utils import SubscribableDict


SOURCE = """
import time
time.sleep(0.001)
""".strip()


@pytest.fixture()
def registry():
    y = SubscribableDict()
    y["run_no_count"] = itertools.count().__next__
    y["statement"] = SOURCE
    yield y
    y.close()


@pytest.mark.asyncio
async def test_register_state_name(registry):
    async def subscribe():
        return [y async for y in registry.subscribe("state_name")]

    async def run():
        await asyncio.sleep(0)
        state = Initialized(registry=registry)
        state = state.run()
        state = await state.exited()
        state = await state.finish()
        state = state.close()
        registry.close()

    actual, _ = await asyncio.gather(subscribe(), run())

    expected = ["initialized", "running", "exited", "finished", "closed"]
    assert expected == actual


@pytest.mark.asyncio
async def test_register_state_name_reset(registry):
    async def subscribe():
        return [y async for y in registry.subscribe("state_name")]

    async def run():
        await asyncio.sleep(0)
        state = Initialized(registry=registry)
        state = state.reset()
        state = state.run()
        state = await state.exited()
        state = await state.finish()
        state = state.reset()
        state = state.run()
        state = await state.exited()
        state = await state.finish()
        state = state.close()
        registry.close()

    actual, _ = await asyncio.gather(subscribe(), run())

    expected = [
        "initialized",
        "initialized",
        "running",
        "exited",
        "finished",
        "initialized",
        "running",
        "exited",
        "finished",
        "closed",
    ]
    assert expected == actual



import itertools

import pytest

from nextline.state import (
    Initialized,
    Running,
    Exited,
    Finished,
    Closed,
)
from nextline.utils import SubscribableDict

SOURCE = """
import time
time.sleep(0.001)
""".strip()


@pytest.fixture()
def registry():
    y = SubscribableDict()
    y["run_no_count"] = itertools.count().__next__
    yield y
    y.close()


@pytest.mark.asyncio
async def test_transition(registry):
    registry["statement"] = SOURCE

    state = Initialized(registry=registry)
    assert isinstance(state, Initialized)

    state = state.run()
    assert isinstance(state, Running)

    state = await state.exited()
    assert isinstance(state, Exited)

    state = await state.finish()
    assert isinstance(state, Finished)

    state = state.close()
    assert isinstance(state, Closed)



from __future__ import annotations

import queue
from itertools import count
from weakref import WeakKeyDictionary

from ..utils import UniqThreadTaskIdComposer, ThreadTaskDoneCallback
from .ci import PdbCommandInterface
from .custom import CustomizedPdb
from .stream import StreamIn, StreamOut


from typing import Any, Optional, Set, Callable, TYPE_CHECKING, Tuple
from types import FrameType

if TYPE_CHECKING:
    from ..types import TraceFunc
    from ..registry import PdbCIRegistry
    from ..utils import SubscribableDict


def PdbInterfaceFactory(
    registry: SubscribableDict,
    pdb_ci_registry: PdbCIRegistry,
    modules_to_trace: Set[str],
) -> Callable[[], PdbInterface]:

    thread_task_id_composer = UniqThreadTaskIdComposer()
    trace_id_counter = count(1).__next__
    prompting_counter = count(1).__next__
    callback_map: WeakKeyDictionary[Any, PdbInterface] = WeakKeyDictionary()

    def callback_func(key):
        callback_map[key].close()

    callback = ThreadTaskDoneCallback(done=callback_func)

    def factory() -> PdbInterface:
        # TODO: check if already created for the same thread or task
        pbi = PdbInterface(
            trace_id_counter=trace_id_counter,
            thread_task_id_composer=thread_task_id_composer,
            registry=registry,
            ci_registry=pdb_ci_registry,
            prompting_counter=prompting_counter,
            modules_to_trace=modules_to_trace,
        )
        key = callback.register()
        callback_map[key] = pbi
        return pbi

    return factory


class PdbInterface:
    """Instantiate Pdb and register its command loops

    TODO: Update parameters

    Parameters
    ----------
    trace_id : object
        The Id to distiugish each instance of Pdb
    modules_to_trace: set
        The set of modules to trace. This object is shared by multiple
        objects. Modules in which Pdb commands are prompted will be
        added.
    registry: object
        A registry
    ci_registry: object
        A registry
    prompting_counter : callable
        Used to count the Pdb command loops
    """

    def __init__(
        self,
        trace_id_counter: Callable[[], int],
        thread_task_id_composer: UniqThreadTaskIdComposer,
        registry: SubscribableDict,
        ci_registry: PdbCIRegistry,
        prompting_counter: Callable[[], int],
        modules_to_trace: Set[str],
    ):
        self._trace_id_counter = trace_id_counter
        self._thread_task_id_composer = thread_task_id_composer
        self._registry = registry
        self._ci_registry = ci_registry
        self._prompting_counter = prompting_counter
        self.modules_to_trace = modules_to_trace
        self._opened = False

    def open(self) -> TraceFunc:
        self._q_stdin: queue.Queue = queue.Queue()
        self._q_stdout: queue.Queue = queue.Queue()

        self._pdb = CustomizedPdb(
            pdbi=self,
            stdin=StreamIn(self._q_stdin),
            stdout=StreamOut(self._q_stdout),
            readrc=False,
        )

        self._trace_id = self._trace_id_counter()
        self._registry[self._trace_id] = None
        ids = (self._registry.get("trace_ids") or ()) + (self._trace_id,)
        self._registry["trace_ids"] = ids

        self._opened = True

        return self._pdb.trace_dispatch

    def close(self):
        if not self._opened:
            return
        del self._registry[self._trace_id]
        ids = list(self._registry.get("trace_ids"))
        ids.remove(self._trace_id)
        ids = tuple(ids)
        self._registry["trace_ids"] = ids

    def calling_trace(self, frame: FrameType, event: str, arg: Any) -> None:
        self._current_trace_args: Optional[Tuple] = (frame, event, arg)

    def exited_trace(self) -> None:
        self._current_trace_args = None

    def entering_cmdloop(self) -> None:
        if not self._current_trace_args:
            raise RuntimeError("calling_trace() must be called first")

        frame, event, _ = self._current_trace_args

        module_name = frame.f_globals.get("__name__")
        self.modules_to_trace.add(module_name)

        self._state = {
            "prompting": self._prompting_counter(),
            "file_name": self._pdb.canonic(frame.f_code.co_filename),
            "line_no": frame.f_lineno,
            "trace_event": event,
        }

        self._pdb_ci = PdbCommandInterface(
            self._pdb, self._q_stdin, self._q_stdout
        )
        self._pdb_ci.start()

        self._ci_registry.add(self._trace_id, self._pdb_ci)

        copy = self._state.copy()
        self._registry[self._trace_id] = copy

    def exited_cmdloop(self) -> None:
        self._state["prompting"] = 0

        self._ci_registry.remove(self._trace_id)

        copy = self._state.copy()
        self._registry[self._trace_id] = copy

        self._pdb_ci.end()



import asyncio
from typing import AsyncGenerator, Iterable, Any


async def aiterable(iterable: Iterable) -> AsyncGenerator[Any, None]:
    '''Wrap iterable so can be used with "async for"'''
    for i in iterable:
        await asyncio.sleep(0)  # let other tasks run
        yield i



import threading


##__________________________________________________________________||
class PdbCIRegistry:
    """Hold the list of active pdb command interfaces"""

    def __init__(self):
        self.pdb_cis = []
        self._dict = {}
        self.condition = threading.Condition()

    def add(self, thread_task_id, pdb_ci):
        with self.condition:
            self._dict[thread_task_id] = pdb_ci
            self.pdb_cis.append(pdb_ci)

    def remove(self, thread_task_id):
        with self.condition:
            pdb_ci = self._dict.pop(thread_task_id)
            self.pdb_cis.remove(pdb_ci)

    def get_ci(self, thread_task_id):
        with self.condition:
            return self._dict.get(thread_task_id, None)


##__________________________________________________________________||



import pytest

from unittest.mock import Mock

from nextline.registry import PdbCIRegistry


@pytest.fixture(autouse=True)
def monkey_patch_trace(monkeypatch):
    """Mock the class Trace in the module nextline.state"""
    mock_instance = Mock()
    mock_instance.return_value = None
    mock_instance.pdb_ci_registry = Mock(spec=PdbCIRegistry)
    mock_class = Mock(return_value=mock_instance)
    monkeypatch.setattr("nextline.state.Trace", mock_class)
    yield mock_class



from __future__ import annotations

import asyncio
from threading import Thread
import itertools
from typing import Optional, Any

from .trace import Trace
from .registry import PdbCIRegistry
from .utils import SubscribableDict, ThreadSafeAsyncioEvent, to_thread
from .call import call_with_trace
from . import script

SCRIPT_FILE_NAME = "<string>"


# __________________________________________________________________||
class Machine:
    """State machine

                 .-------------.
            .--->| Initialized |---.
    reset() |    '-------------'   |
            |      |   | run()     |
            |------'   |           |
            |          v           |
            |    .-------------.   |
            |    |   Running   |   |
            |    '-------------'   |
            |          | _exited() |
            |          |           |
            |          v           |
            |    .-------------.   |
            |    |   Exited    |   |
            |    '-------------'   |
            |          | finish()  |
            |          |           |
            |          V           |
            |    .-------------.   |  close()  .-------------.
            '----|  Finished   |-------------->|   Closed    |
                 '-------------'               '-------------'


    Parameters
    ----------
    statement : str
        A Python code as a string.

    """

    def __init__(self, statement: str, run_no_start_from=1):
        self.registry = SubscribableDict()

        self.registry["statement"] = statement
        self.registry["script_file_name"] = SCRIPT_FILE_NAME
        self.registry["run_no_count"] = itertools.count(
            run_no_start_from
        ).__next__
        self.registry["thread_task_ids"] = ()

        self._state: State = Initialized(self.registry)

        self._lock_finish = asyncio.Condition()
        self._lock_close = asyncio.Condition()

    def __repr__(self):
        # e.g., "<Machine 'running'>"
        return f"<{self.__class__.__name__} {self.state_name!r}>"

    @property
    def state_name(self) -> str:
        """e.g., "initialized", "running","""
        try:
            return self._state.name
        except BaseException:
            return "unknown"

    def run(self) -> None:
        """Enter the running state"""
        self._state = self._state.run()
        self._task_exited = asyncio.create_task(self._exited())

    async def _exited(self) -> None:
        """Enter the exited state

        This method is scheduled as a task in run().

        Before changing the state to "exited", this method waits for the script
        execution (in another thread) to exit.
        """

        self._state = await self._state.exited()

    def send_pdb_command(self, thread_asynctask_id, command) -> None:
        self._state.send_pdb_command(thread_asynctask_id, command)

    async def finish(self) -> None:
        """Enter the finished state"""
        await self._task_exited
        async with self._lock_finish:
            self._state = await self._state.finish()

    def exception(self) -> Optional[Exception]:
        return self._state.exception()

    def result(self) -> Any:
        return self._state.result()

    def reset(self, statement: Optional[str] = None) -> None:
        """Enter the initialized state"""
        if statement:
            self.registry["statement"] = statement
        self._state = self._state.reset()

    async def close(self) -> None:
        """Enter the closed state"""
        async with self._lock_close:
            self._state = self._state.close()
            await to_thread(self.registry.close)


# __________________________________________________________________||
class StateObsoleteError(Exception):
    """Operation on an obsolete state object."""

    pass


class StateMethodError(Exception):
    """Irrelevant operation on the state."""

    pass


# __________________________________________________________________||
class ObsoleteMixin:
    def assert_not_obsolete(self):
        if self.is_obsolete():
            raise StateObsoleteError(f"The state object is obsolete: {self!r}")

    def is_obsolete(self):
        return getattr(self, "_obsolete", False)

    def obsolete(self):
        self._obsolete = True


class State(ObsoleteMixin):
    """The base state class in the Nextline state machine"""

    name = "state"

    def __repr__(self):
        # e.g., "<Initialized 'initialized'>"
        items = [self.__class__.__name__, repr(self.name)]
        if self.is_obsolete():
            items.append("obsolete")
        return f'<{" ".join(items)}>'

    def run(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def exited(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def finish(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def reset(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def close(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def send_pdb_command(self, *_, **__) -> None:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def exception(self) -> Optional[Exception]:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def result(self) -> Any:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")


class Initialized(State):
    """The state "initialized", ready to run

    Parameters
    ----------
    registry : object
        An instance of Registry.

    """

    name = "initialized"

    def __init__(self, registry: SubscribableDict):
        self.registry = registry
        run_no = self.registry.get("run_no_count")()
        self.registry["run_no"] = run_no
        self.registry["state_name"] = self.name

    def run(self):
        self.assert_not_obsolete()
        running = Running(self.registry)
        self.obsolete()
        return running

    def reset(self):
        self.assert_not_obsolete()
        initialized = Initialized(registry=self.registry)
        self.obsolete()
        return initialized

    def close(self):
        self.assert_not_obsolete()
        closed = Closed(self.registry)
        self.obsolete()
        return closed


class Running(State):
    """The state "running", the script is being executed.

    Parameters
    ----------
    registry : object
        An instance of Registry
    """

    name = "running"

    def __init__(self, registry: SubscribableDict):
        self.registry = registry
        self._event = ThreadSafeAsyncioEvent()

        statement = self.registry.get("statement")

        self.pdb_ci_registry = PdbCIRegistry()
        trace = Trace(
            registry=self.registry,
            pdb_ci_registry=self.pdb_ci_registry,
        )
        # self.pdb_ci_registry = trace.pdb_ci_registry

        self.registry["state_name"] = self.name

        self.loop = asyncio.get_running_loop()

        def run():
            code = statement
            if isinstance(code, str):
                try:
                    code = compile(code, SCRIPT_FILE_NAME, "exec")
                except BaseException as e:
                    self._done(None, e)
                    return
            func = script.compose(code)
            result = None
            exception = None
            try:
                result = call_with_trace(func, trace)
            except BaseException as e:
                exception = e
            self._done(result, exception)

        self._thread = Thread(target=run, daemon=True)
        self._thread.start()

    def _done(self, result=None, exception=None):

        if self.loop.is_closed():
            # The exit is not being waited in the main thread, for example,
            # exited() is not called.
            return

        self._exited = Exited(
            self.registry,
            thread=self._thread,
            result=result,
            exception=exception,
        )

        try:
            self._event.set()
        except RuntimeError:
            # The event loop is closed.
            pass

    async def exited(self):
        """return the exited state after the script exits."""
        await self._event.wait()
        self.obsolete()
        return self._exited

    def send_pdb_command(self, thread_asynctask_id, command):
        pdb_ci = self.pdb_ci_registry.get_ci(thread_asynctask_id)
        pdb_ci.send_pdb_command(command)


class Exited(State):
    """The state "exited", the script execution has exited

    Parameters
    ----------
    registry : object
        An instance of Registry
    thread : object
        The object of the thread in which the script was executed.
        This thread is to be joined.
    result : any
        The result of the script execution, always None
    exception : exception or None
        The execution raised in the script execution if any. Otherwise
        None
    """

    name = "exited"

    def __init__(self, registry: SubscribableDict, thread, result, exception):
        self.registry = registry
        self._thread = thread
        self._result = result
        self._exception = exception

        self.registry["state_name"] = self.name

    async def finish(self):
        self.assert_not_obsolete()
        await to_thread(self._thread.join)
        finished = Finished(
            self.registry, result=self._result, exception=self._exception
        )
        self.obsolete()
        return finished


class Finished(State):
    """The state "finished", the script execution has finished

    The thread which executed the script has been joined.

    Parameters
    ----------
    registry : object
        An instance of Registry
    result : any
        The result of the script execution, always None
    exception : exception or None
        The exception of the script execution if any. Otherwise None

    """

    name = "finished"

    def __init__(self, registry: SubscribableDict, result, exception):
        self._result = result
        self._exception = exception

        self.registry = registry
        self.registry["state_name"] = self.name

    def exception(self):
        """Return the exception of the script execution

        Return None if no exception has been raised.

        """
        return self._exception

    def result(self):
        """Return the result of the script execution

        None in the current implementation as the build-in function
        exec() returns None.

        Re-raise the exception if an exception has been raised in the
        script execution.

        """

        if self._exception:
            raise self._exception

        return self._result

    async def finish(self):
        # This method can be called when Nextline.finish() is
        # asynchronously called multiple times.
        self.assert_not_obsolete()
        return self

    def reset(self):
        self.assert_not_obsolete()
        initialized = Initialized(registry=self.registry)
        self.obsolete()
        return initialized

    def close(self):
        self.assert_not_obsolete()
        closed = Closed(self.registry)
        self.obsolete()
        return closed


class Closed(State):
    """The state "closed"

    Parameters
    ----------
    registry : object
        An instance of Registry
    """

    name = "closed"

    def __init__(self, registry: SubscribableDict):
        self.registry = registry
        self.registry["state_name"] = self.name

    def close(self):
        self.assert_not_obsolete()
        return self


# __________________________________________________________________||



import asyncio
import threading
from functools import partial

import pytest

from nextline.utils import to_thread
from nextline.utils import UniqThreadTaskIdComposer as IdComposer
from nextline.utils.types import ThreadTaskId


##__________________________________________________________________||
@pytest.fixture(autouse=True)
def wrap_thread(monkeypatch):
    """Replace threading.Thread with nextline.utils.ExcThread"""

    from nextline.utils import ExcThread

    monkeypatch.setattr(threading, "Thread", ExcThread)

    yield


##__________________________________________________________________||
def assert_call(obj: IdComposer, expected: ThreadTaskId):
    assert expected == obj()
    assert expected == obj()


async def async_assert_call(obj: IdComposer, expected: ThreadTaskId):
    await asyncio.sleep(0)
    assert expected == obj()
    await asyncio.sleep(0)
    assert expected == obj()


##__________________________________________________________________||
@pytest.fixture()
def obj():
    y = IdComposer()
    yield y


##__________________________________________________________________||
def test_compose(obj: IdComposer):
    expected = (1, None)
    assert_call(obj, expected)


def test_threads(obj: IdComposer):
    expected = (1, None)
    assert_call(obj, expected)

    expected = (2, None)
    t = threading.Thread(target=assert_call, args=(obj, expected))
    t.start()
    t.join()

    expected = (3, None)
    t = threading.Thread(target=assert_call, args=(obj, expected))
    t.start()
    t.join()


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_async_coroutine(obj: IdComposer):
    expected = (1, 1)
    assert_call(obj, expected)

    # run in the same task
    await async_assert_call(obj, expected)
    await async_assert_call(obj, expected)


@pytest.mark.asyncio
async def test_async_tasks(obj: IdComposer):
    expected = (1, 1)
    assert_call(obj, expected)

    expected = (1, 2)
    t = asyncio.create_task(async_assert_call(obj, expected))
    await t

    expected = (1, 3)
    t = asyncio.create_task(async_assert_call(obj, expected))
    await t


@pytest.mark.asyncio
async def test_async_tasks_gather(obj: IdComposer):
    expected = (1, 1)
    assert_call(obj, expected)

    expected = (1, 2)
    t1 = asyncio.create_task(async_assert_call(obj, expected))
    expected = (1, 3)
    t2 = asyncio.create_task(async_assert_call(obj, expected))
    aws = {t1, t2}
    await asyncio.gather(*aws)


def test_async_asyncio_run(obj: IdComposer):
    expected = (1, None)
    assert_call(obj, expected)

    expected = (1, 1)
    asyncio.run(async_assert_call(obj, expected))


@pytest.mark.asyncio
async def test_async_asyncio_to_thread(obj: IdComposer):
    expected = (1, 1)
    assert_call(obj, expected)

    expected = (2, None)
    await to_thread(partial(assert_call, obj, expected))

    expected = (2, None)
    await to_thread(partial(assert_call, obj, expected))


##__________________________________________________________________||
async def async_nested(obj: IdComposer, expected_thread_id):
    expected1 = (expected_thread_id, 1)
    assert_call(obj, expected1)
    await async_assert_call(obj, expected1)

    expected2 = (expected_thread_id, 2)
    t1 = asyncio.create_task(async_assert_call(obj, expected2))
    expected3 = (expected_thread_id, 3)
    t2 = asyncio.create_task(async_assert_call(obj, expected3))
    aws = {t1, t2}
    await asyncio.gather(*aws)


def nested(obj: IdComposer, expected_thread_id):
    expected = (expected_thread_id, None)
    assert_call(obj, expected)

    asyncio.run(async_nested(obj, expected_thread_id))


def test_nested(obj: IdComposer):
    expected = (1, None)
    assert_call(obj, expected)

    expected_thread_id = 2
    t = threading.Thread(target=nested, args=(obj, expected_thread_id))
    t.start()
    t.join()

    expected_thread_id = 3
    t = threading.Thread(target=nested, args=(obj, expected_thread_id))
    t.start()
    t.join()


##__________________________________________________________________||



import threading


##__________________________________________________________________||
class PdbCommandInterface:
    """Relay pdb command prompts and commands

    An instance of this class is created for each execution of the pdb
    command loop, pdb._cmdloop().

    Parameters
    ----------
    pdb : Pdb
        The Pdb instance executing _cmdloop()
    queue_in : queue
        The queue connected to stdin in pdb
    queue_out : queue
        The queue connected to stdout in pdb

    """

    def __init__(self, pdb, queue_in, queue_out):
        self.pdb = pdb
        self.queue_in = queue_in
        self.queue_out = queue_out
        self.ended = False
        self.nprompts = 0

    def send_pdb_command(self, command):
        """send a command to pdb"""
        self.command = command
        self.queue_in.put(command)

    def start(self):
        """start interfacing the pdb"""
        self.thread = threading.Thread(target=self._receive_pdb_stdout)
        self.thread.start()

    def end(self):
        """end interfacing the pdb"""
        self.ended = True
        self.queue_out.put(None)  # end the thread
        self.thread.join()

    def _receive_pdb_stdout(self):
        """receive stdout from pdb

        This method runs in its own thread during pdb._cmdloop()
        """
        while out := self._read_uptp_prompt(self.queue_out, self.pdb.prompt):
            self.nprompts += 1
            self.stdout = out

    def _read_uptp_prompt(self, queue, prompt):
        """read the queue up to the prompt"""
        out = ""
        while True:
            m = queue.get()
            if m is None:  # end
                return None
            out += m
            if prompt == m:
                break
        return out


##__________________________________________________________________||



import asyncio
from asyncio import Task
from threading import Thread

import pytest
from unittest.mock import Mock

from typing import Union, Callable, Dict, Set

from nextline.trace import TraceDispatchThreadOrTask
from nextline.utils import current_task_or_thread
from nextline.types import TraceFunc
from nextline.utils import ExcThread

from .funcs import TraceSummary, summarize_trace_calls

from . import module_a, module_b


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
    factory: Union[Mock, Callable[[], Union[Mock, TraceFunc]]],
    probes: Dict[Union[Task, Thread], TraceSummary],
    probe_trace_funcs: Dict[Union[Task, Thread], Union[Mock, TraceFunc]],
    task_or_threads: Dict[Union[Task, Thread], Set[Union[Task, Thread]]],
    modules_in_summary: Set[str],
):
    assert modules_in_summary is not None
    assert ref.call.module
    assert ref.return_.module
    assert ref.call.func == target.call.func
    assert not target.return_.func

    assert factory.call_count == len(task_or_threads)
    assert probe_trace_funcs.keys() == task_or_threads.keys()
    assert probes.keys() == task_or_threads.keys()

    for created_in, called_in in task_or_threads.items():
        assert {created_in} == called_in

    assert ref.call.func == probe.call.func
    assert ref.return_.func == probe.return_.func


@pytest.fixture()
def modules_in_summary():
    yield {__name__, module_a.__name__, module_b.__name__}


def f1():
    module_a.func_a()


def f2():
    t = ExcThread(target=module_a.func_a)
    t.start()
    t.join()


def f3():
    async def a():
        module_a.func_a()

    asyncio.run(a())


def f4():
    t = ExcThread(target=f3)
    t.start()
    t.join()


@pytest.fixture(params=[f1, f2, f3, f4])
def func(request):
    yield request.param


@pytest.fixture()
def target_trace_func(factory: Mock):
    y = TraceDispatchThreadOrTask(factory=factory)
    yield y


@pytest.fixture(params=[True, False])
def thread(request):
    y = request.param
    yield y


@pytest.fixture()
def probes(
    probe_trace_funcs: Dict[Union[Task, Thread], Union[Mock, TraceFunc]],
    modules_in_summary: Union[Set[str], None],
):
    y = {
        k: summarize_trace_calls(v, modules=modules_in_summary)
        for k, v in probe_trace_funcs.items()
    }
    yield y


@pytest.fixture()
def factory(probe_trace_funcs, task_or_threads, probe_trace_func):
    def factory_():
        called_in = set()

        wrap_trace = Mock(wraps=probe_trace_func)

        def trace(*a, **k):
            called_in.add(current_task_or_thread())
            local_trace = probe_trace_func(*a, **k)
            if local_trace is probe_trace_func:
                return wrap_trace
            return local_trace

        wrap_trace.side_effect = trace

        created_in = current_task_or_thread()
        task_or_threads[created_in] = called_in
        probe_trace_funcs[created_in] = wrap_trace
        return wrap_trace

    yield Mock(wraps=factory_)


@pytest.fixture()
def probe_trace_funcs():
    yield {}


@pytest.fixture()
def task_or_threads():
    yield {}



import asyncio
import time
from typing import Iterable

import pytest

from nextline.utils import QueueDist, ThreadSafeAsyncioEvent, to_thread

from .aiterable import aiterable


def test_daemon():
    _ = QueueDist()
    # The program shouldn't be blocked even when close() is not called.


@pytest.fixture()
def obj():
    y = QueueDist()
    yield y
    y.close()


def test_open_close(obj):
    assert obj


def test_close_multiple_times(obj):
    obj.close()
    obj.close()


async def async_send(obj: QueueDist, items: Iterable):
    async for i in aiterable(items):
        obj.put(i)
        assert i == obj.get()


async def async_receive(obj: QueueDist):
    return [i async for i in obj.subscribe()]


@pytest.mark.asyncio
async def test_subscribe(obj: QueueDist):
    items = list(range(10))
    task_receive_1 = asyncio.create_task(async_receive(obj))
    task_receive_2 = asyncio.create_task(async_receive(obj))
    task_send = asyncio.create_task(async_send(obj, items))
    await task_send
    results = await asyncio.gather(
        task_receive_1, task_receive_2, to_thread(obj.close)
    )
    result1, result2, *_ = results
    assert items == result1
    assert items == result2
    assert obj.nsubscriptions == 0


@pytest.mark.asyncio
async def test_receive_the_most_recent_item(obj: QueueDist):

    task_receive_1 = asyncio.create_task(async_receive(obj))
    await asyncio.sleep(0)  # let task_receive_1 start

    pre_items = ["A", "B", "C"]
    for i in pre_items:
        obj.put(i)
        assert i == obj.get()

    time.sleep(0.01)

    task_receive_2 = asyncio.create_task(async_receive(obj))

    items = list(range(10))
    task_send = asyncio.create_task(async_send(obj, items))

    await task_send
    obj.close()
    assert [*pre_items, *items] == await task_receive_1

    # receive 'C', which was put before task_receive_2 started
    assert [pre_items[-1], *items] == await task_receive_2


@pytest.mark.asyncio
async def test_subscribe_after_end(obj: QueueDist):
    # Note: This test might be unnecessary. It is more useful to test
    # what happens if subscribed after closed.

    items = list(range(10))
    task_send = asyncio.create_task(async_send(obj, items))
    await task_send
    obj.close()

    task_receive = asyncio.create_task(async_receive(obj))
    await task_receive


async def async_receive_with_break(obj, at=None):
    ret = []
    async for i in obj.subscribe():
        if i == at:
            break
        ret.append(i)
    return ret


@pytest.mark.asyncio
async def test_break(obj: QueueDist):
    items = list(range(10))
    at = 5
    task_receive_1 = asyncio.create_task(async_receive_with_break(obj, at))
    task_receive_2 = asyncio.create_task(async_receive(obj))
    task_send = asyncio.create_task(async_send(obj, items))
    await task_send
    assert items[: items.index(at)] == await task_receive_1
    results = await asyncio.gather(task_receive_2, to_thread(obj.close))
    result2, *_ = results
    assert items == result2
    assert obj.nsubscriptions == 0


nsubscribers = [0, 1, 2, 5, 50]
pre_nitems = [0, 1, 2, 50]
post_nitems = [0, 1, 2, 100]


@pytest.mark.parametrize("post_nitems", post_nitems)
@pytest.mark.parametrize("pre_nitems", pre_nitems)
@pytest.mark.parametrize("nsubscribers", nsubscribers)
@pytest.mark.asyncio
async def test_thread(
    obj: QueueDist,
    nsubscribers: int,
    pre_nitems: int,
    post_nitems: int,
):
    """test if the issue is resolved
    https://github.com/simonsobs/nextline/issues/2

    """

    nitems = pre_nitems + post_nitems

    if nsubscribers >= 50 and nitems >= 100:
        pytest.skip("nsubscribers >= 50 and nitems >= 100")

    async def subscribe(obj, event):
        await event.wait()
        items = []
        async for i in obj.subscribe():
            items.append(i)
        return items

    def send(obj, pre_items, event, post_items, event_end):
        for i in pre_items:
            obj.put(i)
            assert i == obj.get()
        event.set()
        for i in post_items:
            obj.put(i)
            assert i == obj.get()
        event_end.set()

    async def close(obj, event_end):
        await event_end.wait()
        obj.close()

    items = list(range(nitems))
    pre_items = items[:pre_nitems]
    post_items = items[pre_nitems:]

    event = ThreadSafeAsyncioEvent()
    event_end = ThreadSafeAsyncioEvent()

    coro = to_thread(send, obj, pre_items, event, post_items, event_end)
    task_send = asyncio.create_task(coro)

    tasks_subscribe = []
    for i in range(nsubscribers):
        coro = subscribe(obj, event)
        task = asyncio.create_task(coro)
        tasks_subscribe.append(task)

    coro = close(obj, event_end)
    task_close = asyncio.create_task(coro)

    results = await asyncio.gather(*tasks_subscribe, task_send, task_close)
    for actual in results[:nsubscribers]:
        if not actual:  # can be empty
            continue
        # print(actual[0])
        expected = items[items.index(actual[0]) :]  # no missing or duplicate
        # items from the first
        # received item
        assert actual == expected



import pytest
from unittest.mock import Mock

from typing import Union

from nextline.trace import TraceCallPdb
from nextline.pdb.proxy import PdbInterface

from .funcs import TraceSummary, summarize_trace_calls

from . import module_a


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
    mock_pdbi: Union[Mock, PdbInterface],
):
    assert ref.call.module
    assert ref.return_.module
    assert ref.call == target.call
    assert not target.return_.module

    assert ref == probe

    assert 1 == mock_pdbi.open.call_count

    assert probe == summarize_trace_calls(mock_pdbi.calling_trace)
    assert (
        mock_pdbi.exited_trace.call_count == mock_pdbi.calling_trace.call_count
    )


def f():
    module_a.func_a()


@pytest.fixture()
def func():
    yield f


@pytest.fixture(params=[set(), {module_a.__name__}])
def modules_to_skip(request):
    y = request.param
    yield y


@pytest.fixture()
def target_trace_func(mock_pdbi: Union[Mock, PdbInterface]):
    y = TraceCallPdb(pdbi=mock_pdbi)
    yield y


@pytest.fixture()
def mock_pdbi(probe_trace_func):
    y = Mock(spec=PdbInterface)
    y.open.return_value = probe_trace_func
    yield y


@pytest.fixture()
def thread():
    yield False



from threading import Thread, current_thread
from asyncio import Task, current_task
from itertools import count
from collections import defaultdict
from weakref import WeakKeyDictionary

from typing import Union, Callable, Tuple, Dict, DefaultDict

from .types import ThreadID, TaskId, ThreadTaskId


class UniqThreadTaskIdComposer:
    """Compose paris of unique thread Id and async task Id"""

    def __init__(self):

        self.thread_id_counter = count(1).__next__

        self._map: Dict[
            Union[Thread, Task], ThreadTaskId
        ] = WeakKeyDictionary()

        self._thread_id_map: Dict[Thread, ThreadID] = WeakKeyDictionary()
        self._task_id_map: Dict[Task, TaskId] = WeakKeyDictionary()

        self._task_id_counter_map: DefaultDict[
            ThreadID, Callable[[], int]
        ] = defaultdict(lambda: count(1).__next__)

    def __call__(self) -> ThreadTaskId:
        """Return the pair of the current thread ID and async task ID

        Returns
        -------
        tuple
            The pair of the current thread ID and async task ID. If
            not in an async task, the async task ID will be None.
        """

        thread, task = self._current_thread_task()

        key = task or thread

        if ret := self._map.get(key):
            return ret

        ret = self._compose(thread, task)

        self._map[key] = ret

        return ret

    def _current_thread_task(self) -> Tuple[Thread, Union[Task, None]]:
        try:
            task = current_task()
        except RuntimeError:
            task = None
        return current_thread(), task

    def _compose(
        self, thread: Thread, task: Union[Task, None]
    ) -> ThreadTaskId:

        thread_id = self._thread_id_map.get(thread)
        if not thread_id:
            thread_id = self.thread_id_counter()
            self._thread_id_map[thread] = thread_id

        if not task:
            return thread_id, None

        task_id = self._task_id_map.get(task)
        if not task_id:
            task_id = self._task_id_counter_map[thread_id]()
            self._task_id_map[task] = task_id

        return thread_id, task_id



from nextline.registry import PdbCIRegistry


def test_monkey_patch_trace(monkey_patch_trace):
    MockTrace = monkey_patch_trace
    trace = MockTrace()
    assert trace() is None
    assert isinstance(trace.pdb_ci_registry, PdbCIRegistry)



from typing import Callable, Any, Optional
from types import FrameType


TraceFunc = Callable[
    [FrameType, str, Any], Optional[Callable[[FrameType, str, Any], Any]]
]
# Copied from (because not sure how to import)
# https://github.com/python/typeshed/blob/b88a6f19cdcf/stdlib/sys.pyi#L245



import pytest

from nextline.state import Exited, Finished

from .base import BaseTestState


class TestExited(BaseTestState):

    state_class = Exited

    @pytest.fixture()
    def state(self, exited):
        yield exited

    @pytest.mark.asyncio
    async def test_finish(self, state):
        finished = await state.finish()
        assert isinstance(finished, Finished)
        await self.assert_obsolete(state)



import sys
from itertools import count

from typing import Callable, Union

import pytest
from unittest.mock import Mock

from nextline.utils import SubscribableDict
from nextline.pdb.proxy import PdbInterface
from nextline.pdb.custom import CustomizedPdb
from nextline.trace import TraceCallPdb


class MockCustomizedPdb(CustomizedPdb):
    """Skip the command loop"""

    def trace_dispatch(self, frame, event, arg):
        return super().trace_dispatch(frame, event, arg)

    def _cmdloop(self):
        self._proxy.entering_cmdloop()
        # super()._cmdloop()
        self._proxy.exited_cmdloop()


@pytest.fixture(autouse=True)
def monkeypatch_customized_pdb(monkeypatch):
    monkeypatch.setattr("nextline.pdb.proxy.CustomizedPdb", MockCustomizedPdb)


@pytest.fixture()
def mock_registrar():
    y = Mock(spec=PdbInterface)
    yield y


@pytest.fixture()
def proxy(mock_registrar: PdbInterface):
    modules_to_trace = {"tests.pdb.subject", __name__}

    y = TraceCallPdb(
        pdbi=mock_registrar,
        # modules_to_trace=modules_to_trace,
    )

    yield y


def func_c():
    return


def func_b():
    func_c()


def func_a():
    func_b()
    return


params = [
    pytest.param(func_a, id="simple"),
]


@pytest.mark.parametrize("subject", params)
def test_proxy(
    proxy: TraceCallPdb,
    mock_registrar: Union[Mock, PdbInterface],
    subject: Callable,
):

    trace_org = sys.gettrace()
    sys.settrace(proxy)
    subject()
    sys.settrace(trace_org)

    # proxy.close()

    # assert mock_registrar.register.called
    # # print(mock_registrar.register.call_args_list)
    # # TODO: Test contents

    assert 1 == mock_registrar.open.call_count
    # assert 1 == mock_registrar.close.call_count
    # assert mock_registrar.entering_cmdloop.called
    # assert mock_registrar.exited_cmdloop.called
    # print(mock_registrar.method_calls)


##__________________________________________________________________||



from __future__ import annotations

from weakref import WeakKeyDictionary

import fnmatch

from typing import Any, Set, Optional, Union, Callable, TYPE_CHECKING
from types import FrameType

from .pdb.proxy import PdbInterfaceFactory
from .utils import current_task_or_thread

if TYPE_CHECKING:
    from .types import TraceFunc
    from .utils import SubscribableDict
    from .pdb.proxy import PdbInterface
    from .registry import PdbCIRegistry


MODULES_TO_SKIP = {
    "threading",
    "queue",
    "importlib",
    "asyncio.*",
    "janus",
    "codec",
    "concurrent.futures.*",
    "selectors",
    "weakref",
    "_weakrefset",
    "socket",
    "logging",
    "os",
    "collections.*",
    "importlib.*",
    "pathlib",
    "typing",
    "posixpath",
    "fnmatch",
    "_pytest.*",
    "pluggy.*",
    "nextline.pdb.*",
    "nextline.utils.*",
    "nextline.queuedist",
    "nextlinegraphql.schema.bindables",
}


def Trace(
    registry: SubscribableDict,
    pdb_ci_registry: PdbCIRegistry,
    modules_to_trace: Optional[Set[str]] = None,
) -> TraceFunc:
    """Create the main trace function


    Parameters
    ----------
    registry : object
        An instance of Registry
    modules_to_trace : set, optional
        The names of modules to trace. The module in which the trace
        is first time called will be always traced even if not in the
        set.

    """

    if modules_to_trace is None:
        modules_to_trace = set()

    modules_to_trace = set(modules_to_trace)
    # Make a copy so that the original won't be modified.
    # modules_to_trace will be shared and modified by
    # multiple objects.

    create_pdbi = PdbInterfaceFactory(
        registry=registry,
        pdb_ci_registry=pdb_ci_registry,
        modules_to_trace=modules_to_trace,
    )

    def create_trace_for_single_thread_or_task():
        """To be called in the thread or task to be traced"""
        pdbi = create_pdbi()
        trace_call_pdb = TraceCallPdb(pdbi=pdbi)
        return TraceSelectFirstModule(
            trace=trace_call_pdb,
            modules_to_trace=modules_to_trace,
        )

    return TraceSkipModule(
        skip=MODULES_TO_SKIP,
        trace=TraceSkipLambda(
            trace=TraceAddFirstModule(
                modules_to_trace=modules_to_trace,
                trace=TraceDispatchThreadOrTask(
                    factory=create_trace_for_single_thread_or_task
                ),
            ),
        ),
    )


def TraceSkipModule(trace: TraceFunc, skip: Set[str]) -> TraceFunc:
    def ret(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        module_name = frame.f_globals.get("__name__")
        if _is_matched_to_any(module_name, skip):
            return None
        return trace(frame, event, arg)

    return ret


def TraceSkipLambda(trace: TraceFunc) -> TraceFunc:
    def ret(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        func_name = frame.f_code.co_name
        if func_name == "<lambda>":
            return None
        return trace(frame, event, arg)

    return ret


def TraceAddFirstModule(
    trace: TraceFunc, modules_to_trace: Set[str]
) -> TraceFunc:
    first = True

    def global_trace(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        if not first:
            return trace(frame, event, arg)

        def create_local_trace():
            next_trace: Union[TraceFunc, None] = trace

            def local_trace(frame, event, arg):
                nonlocal first, next_trace

                if module_name := frame.f_globals.get("__name__"):
                    first = False
                    modules_to_trace.add(module_name)
                    return trace(frame, event, arg)

                next_trace = next_trace(frame, event, arg)
                return local_trace

            return local_trace

        return create_local_trace()(frame, event, arg)

    return global_trace


def TraceDispatchThreadOrTask(factory: Callable[[], TraceFunc]) -> TraceFunc:
    """Create a trace that creates a new trace for each thread or asyncio task"""

    map: WeakKeyDictionary[Any, TraceFunc] = WeakKeyDictionary()

    def ret(frame, event, arg) -> Optional[TraceFunc]:
        key = current_task_or_thread()
        trace = map.get(key)
        if not trace:
            trace = factory()
            map[key] = trace
        return trace(frame, event, arg)

    return ret


def TraceSelectFirstModule(
    trace: TraceFunc,
    modules_to_trace: Set[str],
) -> TraceFunc:
    first = True

    def ret(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        nonlocal first
        if first:
            module_name = frame.f_globals.get("__name__")
            if not _is_matched_to_any(module_name, modules_to_trace):
                return None
            first = False
        return trace(frame, event, arg)

    return ret


def TraceCallPdb(pdbi: PdbInterface) -> TraceFunc:

    pdb_trace: Union[TraceFunc, None] = None
    # Instead of calling pdbi.open() here, set initially None so that
    # pdbi.open() won't be called unless global_trace() is actually
    # called.

    def global_trace(frame, event, arg) -> Optional[TraceFunc]:
        nonlocal pdb_trace
        if not pdb_trace:
            pdb_trace = pdbi.open()  # Bdb.trace_dispatch

        def create_local_trace():
            next_trace: Union[TraceFunc, None] = pdb_trace

            def local_trace(frame, event, arg):
                nonlocal next_trace
                pdbi.calling_trace(frame, event, arg)
                next_trace = next_trace(frame, event, arg)
                pdbi.exited_trace()
                if next_trace:
                    return local_trace

            return local_trace

        return create_local_trace()(frame, event, arg)

    return global_trace


def _is_matched_to_any(word: Union[str, None], patterns: Set[str]) -> bool:
    """Test if the word matches any of the patterns

    This function is based on Bdb.is_skipped_module():
    https://github.com/python/cpython/blob/v3.9.5/Lib/bdb.py#L191
    """
    if word is None:
        return False
    for pattern in patterns:
        if fnmatch.fnmatch(word, pattern):
            return True
    return False



import pytest
from unittest.mock import Mock

from typing import Set

from nextline.trace import Trace
from nextline.pdb.proxy import PdbInterfaceFactory, PdbInterface
from nextline.utils import SubscribableDict
from nextline.registry import PdbCIRegistry

from .funcs import TraceSummary

from . import module_a


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
):
    assert ref.call.module
    assert ref.return_.module
    assert ref.call == target.call
    assert not target.return_.module

    assert ref == probe


def f():
    module_a.func_a()


@pytest.fixture()
def func():
    yield f


@pytest.fixture()
def thread():
    yield False


@pytest.fixture()
def target_trace_func(modules_to_trace: Set[str]):
    y = Trace(
        registry=Mock(spec=SubscribableDict),
        pdb_ci_registry=Mock(spec=PdbCIRegistry),
        modules_to_trace=modules_to_trace,
    )
    yield y


@pytest.fixture()
def modules_to_trace():
    y = set()
    yield y


@pytest.fixture()
def mock_pdbi(probe_trace_func):
    y = Mock(spec=PdbInterface)
    y.open.return_value = probe_trace_func
    yield y


@pytest.fixture(autouse=True)
def MockPdbInterfaceFactory(mock_pdbi, monkeypatch):
    mock_create_pdbi = Mock(return_value=mock_pdbi)
    y = Mock(spec=PdbInterfaceFactory, return_value=mock_create_pdbi)
    monkeypatch.setattr("nextline.trace.PdbInterfaceFactory", y)
    yield y



import linecache

from typing import Any, AsyncGenerator, Optional, Tuple

from .state import Machine


class Nextline:
    """Nextline allows line-by-line execution of concurrent Python scripts

    Nextline supports concurrency with threading and asyncio. It uses multiple
    instances of Pdb, one for each thread and async task.

    Note
    ----
    The running asyncio event loop must exists when Nextline is instantiated.

    Parameters
    ----------
    statement : str
        A Python code as a string

    """

    def __init__(self, statement):
        self.machine = Machine(statement)
        self.registry = self.machine.registry

    def __repr__(self):
        # e.g., "<Nextline 'running'>"
        return f"<{self.__class__.__name__} {self.state!r}>"

    def run(self) -> None:
        """Execute the script"""
        self.machine.run()

    async def finish(self) -> None:
        """Wait until the script execution exits"""
        await self.machine.finish()

    def exception(self) -> Optional[Exception]:
        """Uncaught exception from the last run"""
        return self.machine.exception()

    def result(self) -> Any:
        """Return value of the last run. always None"""
        return self.machine.result()

    def reset(self, statement=None) -> None:
        """Prepare for the next run"""
        self.machine.reset(statement=statement)

    async def close(self) -> None:
        """End gracefully"""
        await self.machine.close()

    @property
    def statement(self) -> str:
        """The script"""
        return self.registry.get("statement")

    @property
    def state(self) -> str:
        """The current condition of the script execution.

        The possible values are "initialized", "running", "exited", "finished",
        "closed"
        """
        return self.machine.state_name

    async def subscribe_state(self) -> AsyncGenerator[str, None]:
        # wish to be able to write with "yield from" but not possible
        # https://stackoverflow.com/a/59079548/7309855
        agen = self.registry.subscribe("state_name")
        async for y in agen:
            yield y

    @property
    def run_no(self):
        """The current run number"""
        return self.registry.get("run_no")

    async def subscribe_run_no(self):
        agen = self.registry.subscribe("run_no")
        async for y in agen:
            yield y

    async def subscribe_trace_ids(self) -> AsyncGenerator[Tuple[int], None]:
        agen = self.registry.subscribe("trace_ids")
        async for y in agen:
            yield y

    async def subscribe_trace_state(self, trace_id: int):
        agen = self.registry.subscribe(trace_id)
        async for y in agen:
            if y is None:
                continue
            yield y

    def send_pdb_command(self, trace_id: int, command: str):
        self.machine.send_pdb_command(trace_id, command)

    def get_source(self, file_name=None):
        if not file_name or file_name == self.registry.get("script_file_name"):
            return self.registry.get("statement").split("\n")
        return [e.rstrip() for e in linecache.getlines(file_name)]

    def get_source_line(self, line_no, file_name=None):
        """
        based on linecache.getline()
        https://github.com/python/cpython/blob/v3.9.5/Lib/linecache.py#L26
        """
        lines = self.get_source(file_name)
        if 1 <= line_no <= len(lines):
            return lines[line_no - 1]
        return ""



from .queuedist import QueueDist  # noqa: F401
from .thread_task_id import UniqThreadTaskIdComposer  # noqa: F401
from .thread_safe_event import ThreadSafeAsyncioEvent  # noqa: F401
from .loop import ToLoop  # noqa: F401
from .subscribabledict import SubscribableDict  # noqa: F401
from .thread_exception import ExcThread  # noqa: F401
from .done_callback import (  # noqa: F401
    ThreadTaskDoneCallback,
    ThreadDoneCallback,
    TaskDoneCallback,
)
from .func import (  # noqa: F401
    current_task_or_thread,
    to_thread,
    agen_with_wait,
)
